// <pre>

// Arithmetic evaluation library by Esther Iserovich

import java.io.*;
import java.util.*;
import java.lang.Double;
public class ArithEval {

    public static double eval(String expr) throws IOException{
	String postfix = toPostfix(expr);
	//System.out.println(postfix);
	return evaluate(postfix);

    }

    // converts regular expression to a postfix expression
    private static String toPostfix(String expr) throws IOException{
	if (expr == null)
	    return null;
	StringReader r = new StringReader(expr);
	StreamTokenizer inStream = new StreamTokenizer(r);
	inStream.ordinaryChar('/'); // recognizes the '/' character
	inStream.ordinaryChar('-'); // recognizes the '-' character
	String output = "";
	Stack stack = new Stack();
	int tokenType = 0;
	int prevToken = -1;
	// loop through the tokens
	do {
	    tokenType = inStream.nextToken();
	   
	     // numbers get concatenated to the output String
	    if (tokenType == StreamTokenizer.TT_NUMBER){
		//System.out.println("token is a #");
		// if two numbers without an operator in between, it's invalid
		if (prevToken == StreamTokenizer.TT_NUMBER){
		    //System.out.println (" Invalid expression. More operators required.");
		    throw new IOException("More operators required");
		}
	   	output += inStream.nval + " ";}
	    // expressions get outputed, or stacked, depending on precedence
	    else if (tokenType == '+' || tokenType == '-' ||tokenType == '*'||
		tokenType == '/' || tokenType == '%' ||tokenType == '('){
		//checks if there are two operators in a row.  If yes, invalid.
		if (tokenType != '(')
		    if (prevToken != StreamTokenizer.TT_NUMBER && 
			prevToken != ')'){
			// handle unary '-'
			if (tokenType == '-')
			    output += '0' + " ";
			else {
				//System.out.println ("Invalid expression.  More operands required.");
				throw new IOException("More operands required");
			}
		    }
		// all preceding operators of greater or equal precedence get
		// outputed, except for the '('.  Then the operator is pushed.
		while (stack.empty() == false && preced(tokenType)<= 
		       preced(((Integer)stack.peek()).intValue()) &&
		       ((Integer)stack.peek()).intValue() != '(') 
		    output += (char)((Integer)stack.pop()).intValue() + " ";
		stack.push(new Integer(tokenType));
		//System.out.println("pushed " + (char)tokenType);
	    }
	    
	    else if (tokenType == ')'){ // output operands inside ()
		    try{// will throw exception if unbalanced ()
			while (((Integer)stack.peek()).intValue() != '(')
			    output += (char)((Integer)stack.pop()).intValue() +
				" ";
			stack.pop();//get rid of the matching '('
		    }
		    //only reached if unbalanced ()
		    catch (EmptyStackException e){
			//System.out.println("Error: Unbalanced parentheses. ");
			throw new IOException("Unbalanced parens");
		    }	
		} 
	    
	    // if there is an invalid char in the expression
	    else if (tokenType == StreamTokenizer.TT_WORD) {
		//System.out.println("Error:Invalid Argument = " +inStream.sval);
		throw new IOException("Invalid argument = "+inStream.sval);
	    }
	    // if there is any other invalid input
	    else if (tokenType != -1) 
	      { //System.out.println("Found invalid argument:" + (char)tokenType);
			throw new IOException("Found invalid argument:"+(char)tokenType);
		  }
	    prevToken = tokenType;
	} while (tokenType != StreamTokenizer.TT_EOF);
	//add whichever operators are left on the stack to the output
	while (stack.empty() == false)
	    output += (char)((Integer)stack.pop()).intValue() + " ";
	//	System.out.println(output);
	return output;
    }

    // method to evaluate postfix expressions
    private static double evaluate(String str) throws IOException{
	if (str == null || str.equals("")== true)
	    return 0;
	Stack stack = new Stack();
	StringReader r = new StringReader(str);
	StreamTokenizer inStream = new StreamTokenizer(r);
	inStream.ordinaryChar('/');
	inStream.ordinaryChar('-');
	//String output = "";
	int tokenType = 0;
	double d1, d2, d3;
	try{ // if there are too many operators, should throw an exception
	do { // loops through the tokens
	    tokenType = inStream.nextToken(); 
	    // numbers get pushed
	    if (tokenType == StreamTokenizer.TT_NUMBER)
	   	stack.push(new Double(inStream.nval)) ;
	    // operators are applied to the top two numbers of the stack
	    // and the answer is pushed back onto the stack

	    else if (tokenType == '+' ){
		d1 = ((Double)stack.pop()).doubleValue();
		d2 = ((Double)stack.pop()).doubleValue();
		//d3 = d2 +d1;
		stack.push(new Double(d2 + d1));
	    }
	    else if (tokenType == '-' ){
		d1 = ((Double)stack.pop()).doubleValue();
		d2 = ((Double)stack.pop()).doubleValue();
		stack.push(new Double(d2 - d1));
	    }
	    else if (tokenType == '*' ){
		d1 = ((Double)stack.pop()).doubleValue();
		d2 = ((Double)stack.pop()).doubleValue();
		stack.push(new Double(d2 * d1));
	    }
	    else if (tokenType == '/' ){
		d1 = ((Double)stack.pop()).doubleValue();
		d2 = ((Double)stack.pop()).doubleValue();
		stack.push(new Double(d2 / d1));
	    }
	    else if (tokenType == '%' ){
		d1 = ((Double)stack.pop()).doubleValue();
		d2 = ((Double)stack.pop()).doubleValue();
		stack.push(new Double(d2 % d1));
	    }
	    else if (tokenType != -1)
		{
			//System.out.println("Unbalanced parentheses."); 
			throw new IOException("Unbalanced parentheses");
		}
	} while (tokenType != StreamTokenizer.TT_EOF);
	}
	catch (EmptyStackException e){
	    //System.out.println("Error: Too many operators.");
	    throw new IOException("Too many operators");
	}
	double answer =  ((Double)stack.pop()).doubleValue();
	if (stack.empty() == false){
	    //System.out.println("Error: Too many operands.");
	    throw new IOException("Too many operands");
	}
	return answer;
    }

    // determines precedence of operators
    private static int preced(int c){
	int p = 0;
	switch (c){
	case '+' : p = 1; break;
	case '-' : p = 1; break;
	case '*' : p = 2; break;
	case '/' : p = 2; break;
	case '%' : p = 2; break;
	case '(' : p = 3; break;   
	}
	return p;
    }
}    


    
